import dataclasses
import logging
from typing import Dict
from typing import List
from typing import MutableMapping
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class Technique:
    """Technique defined in an Ontology"""

    iri: str  # Internationalized Resource Identifier
    names: Tuple[str]  # Human readable name (first is the perferred one)
    description: str  # Human readable description

    @property
    def primary_name(self) -> str:
        return self.names[0]


@dataclasses.dataclass
class TechniqueMetadata:
    """Set of techniques with associated metadata for file (BLISS scan info)
    and data portal (ICAT dataset metafata)."""

    techniques: Set[Technique]

    def get_scan_metadata(self) -> Optional[Dict[str, Union[List[str], str]]]:
        if self.techniques:
            return self._get_nxnote()

    def get_scan_info(self) -> Dict[str, Dict[str, Union[List[str], str]]]:
        if not self.techniques:
            return dict()
        return {
            "techniques": self._get_nxnote(),
            "scan_meta_categories": ["techniques"],
        }

    def fill_scan_info(self, scan_info: MutableMapping) -> None:
        if not self.techniques:
            return
        scan_meta_categories = scan_info.setdefault("scan_meta_categories", list())
        if "techniques" not in scan_meta_categories:
            scan_meta_categories.append("techniques")
        nxnote = scan_info.get("techniques")
        if nxnote is None:
            nxnote = scan_info["techniques"] = dict()
        self._fill_nxnote(nxnote)

    def _get_nxnote(self) -> Dict[str, Union[List[str], str]]:
        names = list()
        iris = list()
        for technique in sorted(
            self.techniques, key=lambda technique: technique.primary_name
        ):
            names.append(technique.primary_name)
            iris.append(technique.iri)
        return {
            "@NX_class": "NXnote",
            "names": names,
            "iris": iris,
        }

    def _fill_nxnote(self, nxnote: MutableMapping) -> None:
        names = nxnote.get("names", [])
        iris = nxnote.get("iris", [])
        techniques = dict(zip(iris, names))
        for technique in self.techniques:
            techniques[technique.iri] = technique.primary_name
        iris, names = zip(*sorted(techniques.items(), key=lambda tpl: tpl[1]))
        nxnote.update(
            {
                "@NX_class": "NXnote",
                "names": list(names),
                "iris": list(iris),
            }
        )

    def fill_dataset_metadata(self, dataset: MutableMapping) -> None:
        if not self.techniques:
            return
        # Currently handles mutable mappings by only using __getitem__ and __setitem__
        # https://gitlab.esrf.fr/bliss/bliss/-/blob/master/bliss/icat/policy.py
        try:
            definitions = dataset["definition"].split(" ")
        except KeyError:
            definitions = list()
        try:
            pids = dataset["technique_pid"].split(" ")
        except KeyError:
            pids = list()
        techniques = dict(zip(pids, definitions))
        for technique in self.techniques:
            techniques[technique.iri] = technique.primary_name
        for key, value in self._get_icat_metadata(techniques).items():
            try:
                dataset[key] = value
            except KeyError:
                if key == "technique_pid":
                    _logger.warning(
                        "Skip ICAT field 'technique_pid' (requires pyicat-plus>=0.2)"
                    )
                    continue
                raise

    def get_dataset_metadata(self) -> Dict[str, str]:
        if not self.techniques:
            return dict()
        techniques = {
            technique.iri: technique.primary_name for technique in self.techniques
        }
        return self._get_icat_metadata(techniques)

    def _get_icat_metadata(self, techniques: Dict[str, str]) -> Dict[str, str]:
        iris, definitions = zip(*sorted(techniques.items(), key=lambda tpl: tpl[1]))
        return {"technique_pid": " ".join(iris), "definition": " ".join(definitions)}



"""Technique information extracted from Ontologies:

- ESRFET: ESRF experimental techniques.
- PaNET: taxonomy and thesaurus of photon and neutron (PaN) experimental techniques.
"""

import json
import sys
from typing import Any
from typing import Dict
from typing import List

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


def load_technniques(name: str) -> List[Dict[str, Any]]:
    json_file = importlib_resources.files(__package__).joinpath(f"{name}.json")
    with open(json_file, "r") as f:
        return json.load(f)



from functools import lru_cache
from typing import Generator
from typing import List
from typing import Set
from typing import Tuple

from ..db import load_technniques as _load_technniques
from .types import Technique
from .types import TechniqueMetadata


def get_technique_metadata(*names: Tuple[str]) -> TechniqueMetadata:
    """Returns an object that can generate several types of metadata
    associated to the provided technique names."""
    return TechniqueMetadata(techniques=set(_iter_from_names(*names)))


def get_techniques(*names: Tuple[str]) -> Set[Technique]:
    """Returns a set of techniques referenced by the provided technique names."""
    return set(_iter_from_names(*names))


@lru_cache(maxsize=1)
def get_all_techniques() -> List[Technique]:
    """Returns a list of techniques."""
    return [
        Technique(
            iri=techique["iri"],
            names=tuple(techique["names"]),
            description=techique["description"],
        )
        for ontology_name in ["ESRFET"]
        for techique in _load_technniques(ontology_name)
    ]


def _iter_from_names(*names: Tuple[str]) -> Generator[Technique, None, None]:
    techniques = get_all_techniques()
    for name in sorted(set(names)):
        for technique in techniques:
            technique_names = set(map(str.lower, technique.names))
            if name.lower() in technique_names:
                yield technique
                break
        else:
            raise KeyError(f"'{name}' is not a known technique name.")



from ..technique import get_all_techniques
from ..technique import get_techniques


def test_get_all_technique():
    assert get_all_techniques()


def test_get_technique():
    all_techniques = set(get_all_techniques())
    subset = get_techniques("XRF", "XRD")
    assert subset
    assert subset < all_techniques



import pytest

from ..technique import get_technique_metadata


def test_get_dataset_metadata():
    metadata = get_technique_metadata("XAS")
    dataset_metadata = {
        "definition": "XAS",
        "technique_pid": "http://purl.org/pan-science/ESRFET#XAS",
    }

    assert metadata.get_dataset_metadata() == dataset_metadata


def test_fill_dataset_metadata():
    metadata = get_technique_metadata("XAS")
    dataset_metadata = {
        "definition": "XAS",
        "technique_pid": "http://purl.org/pan-science/ESRFET#XAS",
    }

    dataset = {}
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata

    dataset = dict(dataset_metadata)
    metadata.fill_dataset_metadata(dataset_metadata)
    assert dataset == dataset_metadata


def test_get_scan_info():
    metadata = get_technique_metadata("XAS")
    scan_info = {
        "scan_meta_categories": ["techniques"],
        "techniques": {
            "@NX_class": "NXnote",
            "names": ["XAS"],
            "iris": ["http://purl.org/pan-science/ESRFET#XAS"],
        },
    }
    assert metadata.get_scan_info() == scan_info
    assert metadata.get_scan_metadata() == scan_info["techniques"]


def test_fill_scan_info():
    metadata = get_technique_metadata("XAS")
    scan_info = {
        "scan_meta_categories": ["techniques"],
        "techniques": {
            "@NX_class": "NXnote",
            "names": ["XAS"],
            "iris": ["http://purl.org/pan-science/ESRFET#XAS"],
        },
    }

    info = {}
    metadata.fill_scan_info(info)
    assert info == scan_info

    scan_info = {
        "scan_meta_categories": ["techniques", "technique"],
        "techniques": {
            "@NX_class": "NXnote",
            "names": ["XAS", "XRF"],
            "iris": [
                "http://purl.org/pan-science/ESRFET#XAS",
                "http://purl.org/pan-science/ESRFET#XRF",
            ],
        },
    }
    info = {
        "scan_meta_categories": ["techniques", "technique"],
        "techniques": {
            "@NX_class": "NXnote",
            "names": ["XRF"],
            "iris": ["http://purl.org/pan-science/ESRFET#XRF"],
        },
    }
    metadata.fill_scan_info(info)
    assert info == scan_info


def test_wrong_technique_metadata():
    with pytest.raises(KeyError, match="'WRONG' is not a known technique name"):
        get_technique_metadata("XAS", "WRONG")


def test_empty_technique_metadata():
    metadata = get_technique_metadata()
    assert metadata.get_dataset_metadata() == dict()
    assert metadata.get_scan_info() == dict()
    assert metadata.get_scan_metadata() is None



from ..technique import get_technique_metadata


def test_get_dataset_metadata():
    metadata = get_technique_metadata("XRF", "XAS")
    dataset_metadata = {
        "definition": "XAS XRF",
        "technique_pid": " ".join(
            [
                "http://purl.org/pan-science/ESRFET#XAS",
                "http://purl.org/pan-science/ESRFET#XRF",
            ]
        ),
    }

    assert metadata.get_dataset_metadata() == dataset_metadata


def test_fill_dataset_metadata():
    metadata = get_technique_metadata("XRF", "XAS")
    dataset_metadata = {
        "definition": "XAS XRF",
        "technique_pid": " ".join(
            [
                "http://purl.org/pan-science/ESRFET#XAS",
                "http://purl.org/pan-science/ESRFET#XRF",
            ]
        ),
    }

    dataset = dict(dataset_metadata)
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata

    dataset = {
        "definition": "XRF",
        "technique_pid": "http://purl.org/pan-science/ESRFET#XRF",
    }
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata

    dataset = {}
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata

    dataset = {
        "definition": "XRD",
        "technique_pid": "http://purl.org/pan-science/ESRFET#XRD",
    }
    dataset_metadata = {
        "definition": "XAS XRD XRF",
        "technique_pid": " ".join(
            [
                "http://purl.org/pan-science/ESRFET#XAS",
                "http://purl.org/pan-science/ESRFET#XRD",
                "http://purl.org/pan-science/ESRFET#XRF",
            ]
        ),
    }
    metadata.fill_dataset_metadata(dataset)
    assert dataset == dataset_metadata


def test_get_scan_info():
    metadata = get_technique_metadata("XRF", "XAS")
    scan_info = {
        "scan_meta_categories": ["techniques"],
        "techniques": {
            "@NX_class": "NXnote",
            "names": ["XAS", "XRF"],
            "iris": [
                "http://purl.org/pan-science/ESRFET#XAS",
                "http://purl.org/pan-science/ESRFET#XRF",
            ],
        },
    }
    assert metadata.get_scan_info() == scan_info
    assert metadata.get_scan_metadata() == scan_info["techniques"]


def test_fill_scan_info():
    metadata = get_technique_metadata("XRF", "XAS")
    scan_info = {
        "scan_meta_categories": ["technique", "techniques"],
        "techniques": {
            "@NX_class": "NXnote",
            "names": ["XAS", "XRF"],
            "iris": [
                "http://purl.org/pan-science/ESRFET#XAS",
                "http://purl.org/pan-science/ESRFET#XRF",
            ],
        },
    }

    info = {
        "scan_meta_categories": ["technique"],
        "techniques": None,
    }
    metadata.fill_scan_info(info)
    assert info == scan_info


def test_double_technique_metadata():
    metadata = get_technique_metadata("XRF", "XAS", "XRF", "XAS")
    assert len(metadata.techniques) == 2

    dataset_metadata = {
        "definition": "XAS XRF",
        "technique_pid": " ".join(
            [
                "http://purl.org/pan-science/ESRFET#XAS",
                "http://purl.org/pan-science/ESRFET#XRF",
            ]
        ),
    }
    assert metadata.get_dataset_metadata() == dataset_metadata

    scan_info = {
        "scan_meta_categories": ["techniques"],
        "techniques": {
            "@NX_class": "NXnote",
            "names": ["XAS", "XRF"],
            "iris": [
                "http://purl.org/pan-science/ESRFET#XAS",
                "http://purl.org/pan-science/ESRFET#XRF",
            ],
        },
    }
    assert metadata.get_scan_info() == scan_info
    assert metadata.get_scan_metadata() == scan_info["techniques"]
